mod ability_right_side;
mod content;

use std::sync::{Arc, Weak};

use anyhow::Result;
use rpg_components::components::inventory::Inventory;

use self::ability_right_side::AbilityPageRightSide;

use super::{
    CharacterWindow, Page,
    content::Content,
    page_content::{EmptyRightSide, PageContent},
    traits::{PageContentWrapper, RightSide},
};
use crate::*;

pub struct AbilityPage<A: Ability + 'static> {
    close: Weak<Button>,

    hero: Entity,

    grid: Arc<Grid>,

    tooltip: Arc<Grid>,
    content: Arc<Grid>,

    modes: [Box<dyn PageContentWrapper>; 2],

    current_mode: usize,

    right_side: AbilityPageRightSide<A>,
}

impl<A: Ability + 'static> AbilityPage<A> {
    pub fn new(
        world: &mut World,
        hero: Entity,
        reference: Weak<CharacterWindow>,
        close: &Arc<Button>,
    ) -> Result<Self> {
        let gui_handler = world.resources.get_mut_unchecked::<GuiHandler>();
        let grid = Grid::new(gui_handler, 2, 1, false)?;

        let left_base = GuiSnippet::from_str(
            world,
            include_str!("../../resources/abilities/left_side.xml"),
        )?;

        grid.attach(world, left_base.clone(), 0, 0, 1, 1)?;

        Self::setup_content_switch(&left_base, reference.clone())?;

        let tooltip = left_base.element("tooltip")?;
        let content = left_base.element("content")?;

        // abilities
        let ability_mode = PageContent::new(
            Content::new::<_, Self>(world, reference.clone(), {
                let hero = hero.clone();

                move |world| {
                    let hero_object = world.entity(hero)?;
                    let inventory = hero_object.get_component::<Inventory<A>>()?;

                    let data = inventory.iter_books().cloned().collect();

                    Ok(data)
                }
            })?,
            {
                let ui = GuiSnippet::from_str(
                    world,
                    include_str!("../../resources/abilities/ability_tooltip.xml"),
                )?;

                let mut resources = world.resources.multi_mut();
                let gui_handler = resources.get::<GuiHandler>();
                let engine_settings = resources.get::<EngineSettings>();
                let context = resources.get::<Context>();

                let equip: Arc<Label> = ui.element("equip")?;
                equip.set_info_icon(
                    gui_handler,
                    &engine_settings.controller_icon(context, ControllerButton::A)?,
                )?;

                let upgrade: Arc<Label> = ui.element("upgrade")?;
                upgrade.set_info_icon(
                    gui_handler,
                    &engine_settings.controller_icon(context, ControllerButton::Y)?,
                )?;

                let salvage: Arc<Label> = ui.element("salvage")?;
                salvage.set_info_icon(
                    gui_handler,
                    &engine_settings.controller_icon(context, ControllerButton::X)?,
                )?;

                let switch_mode: Arc<Label> = ui.element("switch_mode")?;
                switch_mode.set_info_icon(
                    gui_handler,
                    &engine_settings.controller_icon(context, ControllerButton::LeftStick)?,
                )?;

                ui
            },
            EmptyRightSide,
        );

        // addons
        let addons_mode = PageContent::<A, _>::new(
            Content::new::<_, Self>(world, reference.clone(), {
                let hero = hero.clone();

                move |world| {
                    let hero_object = world.entity(hero)?;
                    let inventory = hero_object.get_component::<Inventory<A>>()?;

                    let data = inventory.iter_addons().cloned().collect();

                    Ok(data)
                }
            })?,
            {
                let ui = GuiSnippet::from_str(
                    world,
                    include_str!("../../resources/abilities/addon_tooltip.xml"),
                )?;

                let mut resources = world.resources.multi_mut();
                let gui_handler = resources.get::<GuiHandler>();
                let engine_settings = resources.get::<EngineSettings>();
                let context = resources.get::<Context>();

                let equip: Arc<Label> = ui.element("socket")?;
                equip.set_info_icon(
                    gui_handler,
                    &engine_settings.controller_icon(context, ControllerButton::A)?,
                )?;

                let salvage: Arc<Label> = ui.element("salvage")?;
                salvage.set_info_icon(
                    gui_handler,
                    &engine_settings.controller_icon(context, ControllerButton::X)?,
                )?;

                let switch_mode: Arc<Label> = ui.element("switch_mode")?;
                switch_mode.set_info_icon(
                    gui_handler,
                    &engine_settings.controller_icon(context, ControllerButton::LeftStick)?,
                )?;

                ui
            },
            EmptyRightSide,
        );

        let right_side = AbilityPageRightSide::new(world, hero, &reference)?;

        Ok(Self {
            close: Arc::downgrade(close),

            hero,

            grid,

            tooltip,
            content,

            modes: [Box::new(ability_mode), Box::new(addons_mode)],

            current_mode: 0,

            right_side,
        })
    }

    fn update_page(&mut self, world: &mut World) -> Result<()> {
        match self.current_mode {
            0 => println!("update ability view"),
            1 => println!("update addon view"),

            _ => unreachable!(),
        }

        let mode = &mut self.modes[self.current_mode];
        mode.content_mut().update(world, self.hero)?;

        self.tooltip
            .attach(world, mode.tooltip().clone(), 0, 0, 1, 1)?;
        self.content
            .attach(world, mode.content_mut().base().clone(), 0, 0, 1, 1)?;

        self.right_side.refresh(world, self.hero)?;
        self.grid
            .attach(world, self.right_side.base().clone(), 1, 0, 1, 1)?;

        Ok(())
    }

    fn setup_content_switch(
        left_base: &GuiSnippet,
        reference: Weak<CharacterWindow>,
    ) -> Result<()> {
        let switch = {
            let reference = reference.clone();

            move |world: &mut World, index| {
                if let Some(menu) = reference.upgrade() {
                    let mut tabs = menu.tabs_mut();
                    let me = tabs.abilities::<A>();

                    if me.current_mode != index {
                        me.current_mode = index;
                        me.update_page(world)?;
                    }
                }

                Ok(())
            }
        };

        let switch_to_abilities = Box::new({
            let switch = switch.clone();

            move |world: &mut World| switch(world, 0)
        });

        let switch_to_addons = Box::new({
            let switch = switch.clone();

            move |world: &mut World| switch(world, 1)
        });

        left_base.set_click_callbacks(vec![
            ("abilities", switch_to_abilities),
            ("addons", switch_to_addons),
        ])
    }
}

impl<A: Ability + 'static> Page for AbilityPage<A> {
    fn enable(&mut self, world: &mut World) -> Result<Arc<Grid>> {
        println!("enable AbilityPage");

        for mode in self.modes.iter_mut() {
            mode.content_mut().refresh(world)?;
        }

        self.update_page(world)?;

        Ok(self.grid.clone())
    }

    fn disable(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn select(&self, world: &mut World) -> Result<()> {
        let mode = &self.modes[self.current_mode];

        mode.content().select(world)?;

        if mode.content().is_empty() {
            if let Some(close) = self.close.upgrade() {
                close.select(world.resources.get_mut::<GuiHandler>())?;
            }
        }

        Ok(())
    }

    fn next_tab(&mut self, world: &mut World) -> Result<()> {
        self.modes[self.current_mode]
            .content_mut()
            .next_tab(world, self.hero)
    }

    fn previous_tab(&mut self, world: &mut World) -> Result<()> {
        self.modes[self.current_mode]
            .content_mut()
            .previous_tab(world, self.hero)
    }

    fn event(&mut self, world: &mut World, button: ControllerButton) -> Result<bool> {
        Ok(match button {
            ControllerButton::LeftStick => {
                self.current_mode = (self.current_mode + 1) % self.modes.len();
                self.update_page(world)?;
                self.select(world)?;
                true
            }

            ControllerButton::RightStick => {
                self.right_side.next_ability(world, self.hero)?;
                true
            }

            _ => false,
        })
    }
}
